from .circuits import (
    Circuits,
    CircuitTerminations,
    VirtualCircuits,
    VirtualCircuitTerminations,
)
from .core import DataSources, Jobs, ObjectChanges
from .dcim import (
    Cables,
    ConsolePorts,
    ConsoleServerPorts,
    Devices,
    DeviceTypes,
    FrontPorts,
    Interfaces,
    PowerFeeds,
    PowerOutlets,
    PowerPorts,
    RackReservations,
    Racks,
    RearPorts,
    Termination,
    VirtualChassis,
)
from .ipam import Aggregates, IpAddresses, Prefixes, VlanGroups, Vlans
from .virtualization import VirtualMachineTypes, VirtualMachines
from .wireless import WirelessLans

CONTENT_TYPE_MAPPER = {
    "circuits.circuit": Circuits,
    "circuits.circuittermination": CircuitTerminations,
    "circuits.virtualcircuit": VirtualCircuits,
    "circuits.virtualcircuittermination": VirtualCircuitTerminations,
    "core.datasource": DataSources,
    "core.job": Jobs,
    "core.objectchange": ObjectChanges,
    "dcim.cable": Cables,
    "dcim.cablebundle": None,
    "dcim.cablepath": None,
    "dcim.cabletermination": Termination,
    "dcim.consoleport": ConsolePorts,
    "dcim.consoleporttemplate": None,
    "dcim.consoleserverport": ConsoleServerPorts,
    "dcim.consoleserverporttemplate": None,
    "dcim.device": Devices,
    "dcim.devicebay": None,
    "dcim.devicebaytemplate": None,
    "dcim.devicerole": None,
    "dcim.devicetype": DeviceTypes,
    "dcim.frontport": FrontPorts,
    "dcim.frontporttemplate": None,
    "dcim.interface": Interfaces,
    "dcim.interfacetemplate": None,
    "dcim.inventoryitem": None,
    "dcim.inventoryitemrole": None,
    "dcim.inventoryitemtemplate": None,
    "dcim.location": None,
    "dcim.manufacturer": None,
    "dcim.module": None,
    "dcim.modulebay": None,
    "dcim.modulebaytemplate": None,
    "dcim.moduletype": None,
    "dcim.platform": None,
    "dcim.powerfeed": PowerFeeds,
    "dcim.poweroutlet": PowerOutlets,
    "dcim.poweroutlettemplate": None,
    "dcim.powerpanel": None,
    "dcim.powerport": PowerPorts,
    "dcim.powerporttemplate": None,
    "dcim.rack": Racks,
    "dcim.rackgroup": None,
    "dcim.rackreservation": RackReservations,
    "dcim.rackrole": None,
    "dcim.rearport": RearPorts,
    "dcim.rearporttemplate": None,
    "dcim.region": None,
    "dcim.site": None,
    "dcim.sitegroup": None,
    "dcim.virtualchassis": VirtualChassis,
    "extras.configcontext": None,
    "extras.configrevision": None,
    "extras.customfield": None,
    "extras.customlink": None,
    "extras.exporttemplate": None,
    "extras.imageattachment": None,
    "extras.jobresult": None,
    "extras.journalentry": None,
    "extras.report": None,
    "extras.script": None,
    "extras.tag": None,
    "extras.taggeditem": None,
    "extras.webhook": None,
    "ipam.aggregate": Aggregates,
    "ipam.ASN": None,
    "ipam.FHRPgroup": None,
    "ipam.FHRPgroupassignment": None,
    "ipam.IPaddress": IpAddresses,
    "ipam.IPrange": None,
    "ipam.L2VPN": None,
    "ipam.L2VPNtermination": None,
    "ipam.prefix": Prefixes,
    "ipam.RIR": None,
    "ipam.role": None,
    "ipam.routetarget": None,
    "ipam.service": None,
    "ipam.servicetemplate": None,
    "ipam.VLAN": Vlans,
    "ipam.VLANgroup": VlanGroups,
    "ipam.VRF": None,
    "tenancy.contact": None,
    "tenancy.contactassignment": None,
    "tenancy.contactgroup": None,
    "tenancy.contactrole": None,
    "tenancy.tenant": None,
    "tenancy.tenantgroup": None,
    "virtualization.cluster": None,
    "virtualization.clustergroup": None,
    "virtualization.clustertype": None,
    "virtualization.interface": None,
    "virtualization.virtualmachine": VirtualMachines,
    "virtualization.virtualmachinetype": VirtualMachineTypes,
    "wireless.WirelessLAN": WirelessLans,
    "wireless.WirelessLANGroup": None,
    "wireless.wirelesslink": None,
}

TYPE_CONTENT_MAPPER = {v: k for k, v in CONTENT_TYPE_MAPPER.items() if v is not None}



"""
(c) 2017 DigitalOcean

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from pynetbox.core.endpoint import DetailEndpoint
from pynetbox.core.response import JsonField, Record
from pynetbox.models.dcim import Devices
from pynetbox.models.ipam import IpAddresses


class VirtualMachineTypes(Record):
    pass


class VirtualMachines(Record):
    device = Devices
    primary_ip = IpAddresses
    primary_ip4 = IpAddresses
    primary_ip6 = IpAddresses
    config_context = JsonField
    virtual_machine_type = VirtualMachineTypes

    @property
    def render_config(self):
        """
        Represents the ``render-config`` detail endpoint.

        Returns a DetailEndpoint object that is the interface for
        viewing response from the render-config endpoint.

        :returns: :py:class:`.DetailEndpoint`

        :Examples:

        >>> vm = nb.virtualization.virtual_machines.get(123)
        >>> vm.render_config.create()
        """
        return DetailEndpoint(self, "render-config")



from pynetbox.core.api import Api
from pynetbox.core.query import (
    AllocationError,
    ContentError,
    RequestError,
    ParameterValidationError,
)

__version__ = "7.7.0"

# Lowercase alias for backward compatibility
api = Api

__all__ = (
    "Api",
    "AllocationError",
    "ContentError",
    "RequestError",
    "ParameterValidationError",
    "api",
    "__version__",
)



"""
(c) 2017 DigitalOcean

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import contextlib

import requests

from pynetbox.core.app import App, PluginsApp
from pynetbox.core.query import Request, TOKEN_PREFIX
from pynetbox.core.response import Record


class Api:
    """The API object is the point of entry to pynetbox.

    After instantiating the Api() with the appropriate named arguments
    you can specify which app and endpoint you wish to interact with.

    Valid attributes currently are:

    * circuits
    * core (NetBox 3.5+)
    * dcim
    * extras
    * ipam
    * tenancy
    * users
    * virtualization
    * vpn (NetBox 3.7+)
    * wireless

    Calling any of these attributes will return an `App` object which exposes endpoints as attributes.

    ## Additional Attributes

    * **http_session(requests.Session)**: Override the default session with your own. This is used to control
      a number of HTTP behaviors such as SSL verification, custom headers,
      retires, and timeouts.
      See [custom sessions](advanced.md#custom-sessions) for more info.

    ## Parameters

    * **url** (str): The base URL to the instance of NetBox you wish to connect to.
    * **token** (str): Your NetBox token.
    * **threading** (bool, optional): Set to True to use threading in `.all()` and `.filter()` requests.

    ## Raises

    * **AttributeError**: If app doesn't exist.

    ## Examples

    ```python
    import pynetbox
    nb = pynetbox.api(
        'http://localhost:8000',
        token='d6f4e314a5b5fefd164995169f28ae32d987704f'
    )
    list(nb.dcim.devices.all())
    # [test1-leaf1, test1-leaf2, test1-leaf3]
    ```
    """

    def __init__(
        self,
        url,
        token=None,
        threading=False,
        strict_filters=False,
    ):
        """Initialize the API client.

        Args:
            url (str): The base URL to the instance of NetBox you wish to connect to.
            token (str, optional): Your NetBox API token. If not provided, authentication will be required for each request.
            threading (bool, optional): Set to True to use threading in `.all()` and `.filter()` requests, defaults to False.
            strict_filters (bool, optional): Set to True to check GET call filters against OpenAPI specifications (intentionally not done in NetBox API), defaults to False.
        """
        base_url = "{}/api".format(url if url[-1] != "/" else url[:-1])
        self.token = token
        self.base_url = base_url
        self.http_session = requests.Session()
        self.threading = threading
        self.strict_filters = strict_filters

        # Initialize NetBox apps
        self.circuits = App(self, "circuits")
        self.core = App(self, "core")
        self.dcim = App(self, "dcim")
        self.extras = App(self, "extras")
        self.ipam = App(self, "ipam")
        self.tenancy = App(self, "tenancy")
        self.users = App(self, "users")
        self.virtualization = App(self, "virtualization")
        self.vpn = App(self, "vpn")
        self.wireless = App(self, "wireless")
        self.plugins = PluginsApp(self)

    @property
    def version(self):
        """Gets the API version of NetBox.

        Can be used to check the NetBox API version if there are
        version-dependent features or syntaxes in the API.

        ## Returns
        Version number as a string.

        ## Example

        ```python
        import pynetbox
        nb = pynetbox.api(
            'http://localhost:8000',
            token='d6f4e314a5b5fefd164995169f28ae32d987704f'
        )
        nb.version
        # '3.1'
        ```
        """
        version = Request(
            base=self.base_url,
            token=self.token,
            http_session=self.http_session,
        ).get_version()
        return version

    def openapi(self):
        """Returns the OpenAPI spec.

        Quick helper function to pull down the entire OpenAPI spec.
        It is stored in memory to avoid repeated calls on NetBox API.

        ## Returns
        dict: The OpenAPI specification as a dictionary.

        ## Example

        ```python
        import pynetbox
        nb = pynetbox.api(
            'http://localhost:8000',
            token='d6f4e314a5b5fefd164995169f28ae32d987704f'
        )
        nb.openapi()
        # {...}
        ```
        """
        if not (openapi := getattr(self, "_openapi", None)):
            openapi = self._openapi = Request(
                base=self.base_url,
                http_session=self.http_session,
            ).get_openapi()

        return openapi

    def status(self):
        """Gets the status information from NetBox.

        ## Returns
        Dictionary containing NetBox status information.

        ## Raises
        `RequestError`: If the request is not successful.

        ## Example

        ```python
        from pprint import pprint
        pprint(nb.status())
        {
            'django-version': '3.1.3',
            'installed-apps': {
                'cacheops': '5.0.1',
                'debug_toolbar': '3.1.1',
                'django_filters': '2.4.0',
                'django_prometheus': '2.1.0',
                'django_rq': '2.4.0',
                'django_tables2': '2.3.3',
                'drf_yasg': '1.20.0',
                'mptt': '0.11.0',
                'rest_framework': '3.12.2',
                'taggit': '1.3.0',
                'timezone_field': '4.0'
            },
            'netbox-version': '2.10.2',
            'plugins': {},
            'python-version': '3.7.3',
            'rq-workers-running': 1
        }
        ```
        """
        status = Request(
            base=self.base_url,
            token=self.token,
            http_session=self.http_session,
        ).get_status()
        return status

    def create_token(self, username, password):
        """Creates an API token using a valid NetBox username and password.
        Saves the created token automatically in the API object.

        ## Parameters
        * **username** (str): NetBox username
        * **password** (str): NetBox password

        ## Returns
        `Record`: The token as a Record object.

        ## Raises
        `RequestError`: If the request is not successful.

        ## Notes

        NetBox 4.5 introduced v2 tokens. For v2 tokens, `nb.token` is set to
        `nbt_<key>.<token>` (the full auth value required in the Authorization
        header), which differs from `token.key`. For v1 tokens (pre-4.5),
        `nb.token` is the plaintext token value.

        ## Example

        ```python
        import pynetbox
        nb = pynetbox.api("https://netbox-server")
        token = nb.create_token("admin", "netboxpassword")

        # NetBox 4.5+ v2 token: nb.token differs from token.key
        nb.token
        # 'nbt_shortkey1234567.plaintexttoken7890abcdef1234567890abcdef'
        token.key
        # 'shortkey1234567'

        # Pre-4.5 / v1 token: nb.token matches token.key (or token.token)
        nb.token
        # '96d02e13e3f1fdcd8b4c089094c0191dcb045bef'
        ```
        """
        resp = Request(
            base="{}/users/tokens/provision/".format(self.base_url),
            http_session=self.http_session,
        ).post(data={"username": username, "password": password})
        # v2 tokens (NetBox 4.5+): construct auth value as nbt_<key>.<token>
        if resp.get("version") == 2:
            self.token = "{}{}.{}".format(TOKEN_PREFIX, resp["key"], resp["token"])
        else:
            self.token = resp.get("token") or resp["key"]
        return Record(resp, self, None)

    @contextlib.contextmanager
    def activate_branch(self, branch):
        """Context manager to activate the branch by setting the schema ID in the headers.

        **Note**: The NetBox branching plugin must be installed and enabled in your NetBox instance for this functionality to work.

        ## Parameters
        * **branch** (Record): The NetBox branch to activate

        ## Raises
        `ValueError`: If the branch is not a valid NetBox branch.

        ## Example

        ```python
        import pynetbox
        nb = pynetbox.api("https://netbox-server")
        branch = nb.plugins.branching.branches.create(name="testbranch")
        with nb.activate_branch(branch):
            sites = nb.dcim.sites.all()
            # All operations within this block will use the branch's schema
        ```
        """
        if not isinstance(branch, Record) or "schema_id" not in dict(branch):
            raise ValueError(
                f"The specified branch is not a valid NetBox branch: {branch}."
            )

        self.http_session.headers["X-NetBox-Branch"] = branch.schema_id

        try:
            yield
        finally:
            self.http_session.headers.pop("X-NetBox-Branch", None)



"""
(c) 2017 DigitalOcean

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from pynetbox.core.response import JsonField, Record


class DataSources(Record):
    pass


class Jobs(Record):
    pass


class ObjectChanges(Record):
    object_data = JsonField
    postchange_data = JsonField
    prechange_data = JsonField

    def __str__(self):
        return self.request_id



"""
(c) 2017 DigitalOcean

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from pynetbox.core.endpoint import (
    DetailEndpoint,
    RODetailEndpoint,
    ROMultiFormatDetailEndpoint,
)
from pynetbox.core.query import Request
from pynetbox.core.response import JsonField, PathableRecord, Record
from pynetbox.models.circuits import Circuits
from pynetbox.models.ipam import IpAddresses


class TraceableRecord(Record):
    def _build_termination_data(self, termination_list):
        terminations_data = []
        for hop_item_data in termination_list:
            return_obj_class = self._get_obj_class(hop_item_data["url"])
            terminations_data.append(
                return_obj_class(hop_item_data, self.endpoint.api, self.endpoint)
            )

        return terminations_data

    def trace(self):
        req = Request(
            key=str(self.id) + "/trace",
            base=self.endpoint.url,
            token=self.api.token,
            http_session=self.api.http_session,
        ).get()

        ret = []
        for a_terminations_data, cable_data, b_terminations_data in req:
            ret.append(self._build_termination_data(a_terminations_data))
            if not cable_data:
                ret.append(cable_data)
            else:
                return_obj_class = self._get_obj_class(cable_data["url"])
                ret.append(
                    return_obj_class(cable_data, self.endpoint.api, self.endpoint)
                )
            ret.append(self._build_termination_data(b_terminations_data))

        return ret


class DeviceTypes(Record):
    def __str__(self):
        return self.model


class Devices(Record):
    """Devices Object.

    Represents a device response from netbox.

    ## Attributes
    * **primary_ip, ip4, ip6** (list): Tells __init__ in Record() to
        take the `primary_ip` field's value from the API
        response and return an initialized list of IpAddress
        objects
    * **device_type** (obj): Tells __init__ in Record() to take the
        `device_type` field's value from the API response and
        return an initialized DeviceType object
    """

    has_details = True
    device_type = DeviceTypes
    primary_ip = IpAddresses
    primary_ip4 = IpAddresses
    primary_ip6 = IpAddresses
    oob_ip = IpAddresses
    local_context_data = JsonField
    config_context = JsonField

    @property
    def napalm(self):
        """Represents the ``napalm`` detail endpoint.

        Returns a DetailEndpoint object that is the interface for
        viewing response from the napalm endpoint.

        ## Returns
        DetailEndpoint object.

        ## Examples

        ```python
        device = nb.ipam.devices.get(123)
        device.napalm.list(method='get_facts')
        # {"get_facts": {"interface_list": ["ge-0/0/0"]}}
        ```
        """
        return RODetailEndpoint(self, "napalm")

    @property
    def render_config(self):
        """Represents the ``render-config`` detail endpoint.

        Returns a DetailEndpoint object that is the interface for
        viewing response from the render-config endpoint.

        ## Returns
        DetailEndpoint object.

        ## Examples

        ```python
        device = nb.ipam.devices.get(123)
        device.render_config.create()
        ```
        """
        return DetailEndpoint(self, "render-config")


class InterfaceConnections(Record):
    def __str__(self):
        return self.interface_a.name


class InterfaceConnection(Record):
    def __str__(self):
        return self.interface.name


class Interfaces(TraceableRecord):
    interface_connection = InterfaceConnection


class PowerFeeds(TraceableRecord):
    pass


class PowerOutlets(TraceableRecord):
    device = Devices


class PowerPorts(TraceableRecord):
    device = Devices


class ConsolePorts(TraceableRecord):
    device = Devices


class ConsoleServerPorts(TraceableRecord):
    device = Devices


class RackReservations(Record):
    def __str__(self):
        return self.description


class VirtualChassis(Record):
    master = Devices


class RUs(Record):
    device = Devices


class FrontPorts(PathableRecord):
    device = Devices


class RearPorts(PathableRecord):
    device = Devices


class Racks(Record):
    @property
    def units(self):
        """Represents the ``units`` detail endpoint.

        Returns a DetailEndpoint object that is the interface for
        viewing response from the units endpoint.

        ## Returns
        DetailEndpoint object.

        ## Examples

        ```python
        rack = nb.dcim.racks.get(123)
        rack.units.list()
        # {"get_facts": {"interface_list": ["ge-0/0/0"]}}
        ```
        """
        return RODetailEndpoint(self, "units", custom_return=RUs)

    @property
    def elevation(self):
        """Represents the ``elevation`` detail endpoint.

        Returns a multi-format endpoint supporting both JSON and SVG responses.
        The elevation endpoint provides rack unit information and can render
        graphical elevation views.

        ## Returns
        ROMultiFormatDetailEndpoint object supporting JSON and SVG formats.

        ## Examples

        ```python
        rack = nb.dcim.racks.get(123)

        # Get rack units as JSON (list of RU objects)
        rack.elevation.list()

        # Get elevation as SVG diagram
        svg = rack.elevation.list(render='svg')
        ```
        """
        return ROMultiFormatDetailEndpoint(self, "elevation", custom_return=RUs)


class Termination(Record):
    def __str__(self):
        # hacky check to see if we're a circuit termination to
        # avoid another call to NetBox because of a non-existent attr
        # in self.name
        if "circuit" in str(self.url):
            return self.circuit.cid

        return self.name

    device = Devices
    circuit = Circuits


class Cables(Record):
    def __str__(self):
        if len(self.a_terminations) == 1 and len(self.b_terminations) == 1:
            return "{} <> {}".format(self.a_terminations[0], self.b_terminations[0])
        return "Cable #{}".format(self.id)



"""
(c) 2017 DigitalOcean

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import concurrent.futures as cf
import io
import os
import json

from packaging import version

# NetBox v2 token prefix (introduced in NetBox 4.5.0)
TOKEN_PREFIX = "nbt_"


def _is_v2_token(token):
    """Detect if a token is NetBox v2 format.

    V2 tokens (introduced in NetBox 4.5.0) have the format: nbt_<id>.<token>
    The nbt_ prefix is used for secrets detection.

    V1 tokens are simple strings without dots.

    Returns True if token is v2 format, False otherwise.
    """
    if not token or not token.startswith(TOKEN_PREFIX):
        return False

    # Remove nbt_ prefix
    token_body = token[len(TOKEN_PREFIX) :]

    # V2 tokens contain a dot separating the ID from the secret
    return "." in token_body


def _is_file_like(obj):
    if isinstance(obj, (str, bytes)):
        return False
    # Check if it's a standard library IO object OR has a callable read method
    return isinstance(obj, io.IOBase) or (
        hasattr(obj, "read") and callable(getattr(obj, "read"))
    )


def _extract_files(data):
    """Extract file-like objects from data dict.

    Returns a tuple of (clean_data, files) where clean_data has file objects
    removed and files is a dict suitable for requests' files parameter.
    """
    if not isinstance(data, dict):
        return data, None

    files = {}
    clean_data = {}

    for key, value in data.items():
        if _is_file_like(value):
            # Format: (filename, file_obj, content_type)
            # Try to get filename from file object, fallback to key
            filename = getattr(value, "name", None)
            if filename:
                # Extract just the filename, not the full path
                filename = os.path.basename(filename)
            else:
                filename = key
            files[key] = (filename, value)
        elif isinstance(value, tuple) and len(value) >= 2 and _is_file_like(value[1]):
            # Already in (filename, file_obj) or (filename, file_obj, content_type) format
            files[key] = value
        else:
            clean_data[key] = value

    return clean_data, files if files else None


def calc_pages(limit, count):
    """Calculate number of pages required for full results set."""
    return int(count / limit) + (limit % count > 0)


class RequestError(Exception):
    """Basic Request Exception.

    More detailed exception that returns the original requests object
    for inspection. Along with some attributes with specific details
    from the requests object. If return is json we decode and add it
    to the message.

    ## Examples

    ```python
    try:
        nb.dcim.devices.create(name="destined-for-failure")
    except pynetbox.RequestError as e:
        print(e.error)
    ```
    """

    def __init__(self, req):
        if req.status_code == 404:
            self.message = "The requested url: {} could not be found.".format(req.url)
        else:
            try:
                self.message = "The request failed with code {} {}: {}".format(
                    req.status_code, req.reason, req.json()
                )
            except ValueError:
                self.message = (
                    "The request failed with code {} {} but more specific "
                    "details were not returned in json. Check the NetBox Logs "
                    "or investigate this exception's error attribute.".format(
                        req.status_code, req.reason
                    )
                )

        super().__init__(self.message)
        self.req = req
        self.request_body = req.request.body
        self.base = req.url
        self.error = req.text

    def __str__(self):
        return self.message


class AllocationError(Exception):
    """Allocation Exception.

    Used with available-ips/available-prefixes when there is no
    room for allocation and NetBox returns 409 Conflict.
    """

    def __init__(self, req):
        super().__init__(req)
        self.req = req
        self.request_body = req.request.body
        self.base = req.url
        self.error = "The requested allocation could not be fulfilled."

    def __str__(self):
        return self.error


class ContentError(Exception):
    """Content Exception.

    If the API URL does not point to a valid NetBox API, the server may
    return a valid response code, but the content is not json. This
    exception is raised in those cases.
    """

    def __init__(self, req):
        super().__init__(req)
        self.req = req
        self.request_body = req.request.body
        self.base = req.url
        self.error = (
            "The server returned invalid (non-json) data. Maybe not a NetBox server?"
        )

    def __str__(self):
        return self.error


class ParameterValidationError(Exception):
    """API parameter validation Exception.

    Raised when filter parameters do not match Netbox OpenAPI specification.

    ## Examples

    ```python
    try:
        nb.dcim.devices.filter(field_which_does_not_exist="destined-for-failure")
    except pynetbox.ParameterValidationError as e:
        print(e.error)
    ```
    """

    def __init__(self, errors):
        super().__init__(errors)
        self.error = f"The request parameter validation returned an error: {errors}"

    def __str__(self):
        return self.error


class Request:
    """Creates requests to the Netbox API.

    Responsible for building the url and making the HTTP(S) requests to
    Netbox's API.

    ## Parameters

    * **base** (str): Base URL passed in api() instantiation.
    * **filters** (dict, optional): Contains key/value pairs that
        correlate to the filters a given endpoint accepts.
        In (e.g. /api/dcim/devices/?name='test') 'name': 'test'
        would be in the filters dict.
    """

    def __init__(
        self,
        base,
        http_session,
        filters=None,
        limit=None,
        offset=None,
        key=None,
        token=None,
        threading=False,
        expect_json=True,
    ):
        """Instantiates a new Request object.

        ## Parameters

        * **base** (string): Base URL passed in api() instantiation.
        * **filters** (dict, optional): Contains key/value pairs that
            correlate to the filters a given endpoint accepts.
            In (e.g. /api/dcim/devices/?name='test') 'name': 'test'
            would be in the filters dict.
        * **key** (int, optional): Database id of the item being queried.
        * **expect_json** (bool, optional): If True, expects JSON response
            and sets appropriate Accept header. If False, expects raw content
            (e.g., SVG, XML) and returns text. Defaults to True.

        ## Note

        The `count` attribute is not initialized here. It is set dynamically
        by the `get()` method when paginating results, or by `get_count()`
        when explicitly requesting the count. This allows `get_count()` to
        use `hasattr()` to determine if a count has already been fetched.
        """
        self.base = self.normalize_url(base)
        self.filters = filters or None
        self.key = key
        self.token = token
        self.http_session = http_session
        self.url = self.base if not key else "{}{}/".format(self.base, key)
        self.threading = threading
        self.limit = limit
        self.offset = offset
        self.expect_json = expect_json

    def get_openapi(self):
        """Gets the OpenAPI Spec."""
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }

        current_version = version.parse(self.get_version())
        if current_version >= version.parse("3.5"):
            req = self.http_session.get(
                "{}schema/".format(self.normalize_url(self.base)),
                headers=headers,
            )
        else:
            req = self.http_session.get(
                "{}docs/?format=openapi".format(self.normalize_url(self.base)),
                headers=headers,
            )

        if req.ok:
            return req.json()
        else:
            raise RequestError(req)

    def get_version(self):
        """Gets the API version of NetBox.

        Issues a GET request to the base URL to read the API version from the
        response headers.

        ## Returns
        Version number as a string. Empty string if version is not
        present in the headers.

        ## Raises
        RequestError if req.ok returns false.
        """
        headers = {"Content-Type": "application/json"}
        self._add_auth_header(headers)
        req = self.http_session.get(
            self.normalize_url(self.base),
            headers=headers,
        )
        if req.ok or req.status_code == 403:
            return req.headers.get("API-Version", "")
        else:
            raise RequestError(req)

    def get_status(self):
        """Gets the status from /api/status/ endpoint in NetBox.

        ## Returns
        Dictionary as returned by NetBox.

        ## Raises
        RequestError if request is not successful.
        """
        headers = {"Content-Type": "application/json"}
        self._add_auth_header(headers)
        req = self.http_session.get(
            "{}status/".format(self.normalize_url(self.base)),
            headers=headers,
        )
        if req.ok:
            return req.json()
        else:
            raise RequestError(req)

    def normalize_url(self, url):
        """Builds a url for POST actions."""
        if url[-1] != "/":
            return "{}/".format(url)

        return url

    def _add_auth_header(self, headers):
        """Add authorization header to headers dict if token is present.

        ## Parameters
        * **headers** (dict): Headers dictionary to update with authorization.
        """
        if self.token:
            if _is_v2_token(self.token):
                headers["authorization"] = "Bearer {}".format(self.token)
            else:
                headers["authorization"] = "Token {}".format(self.token)

    def _make_call(self, verb="get", url_override=None, add_params=None, data=None):
        # Extract any file-like objects from data
        files = None
        # Verbs that support request bodies with file uploads
        body_verbs = ("post", "put", "patch")

        # Set Accept header based on expected response type
        if self.expect_json:
            headers = {"accept": "application/json"}
        else:
            headers = {"accept": "*/*"}

        # Extract files from data for applicable verbs
        if data is not None and verb in body_verbs:
            data, files = _extract_files(data)

        # Set headers based on request type
        should_be_json_body = not files and (
            verb in body_verbs or (verb == "delete" and data)
        )

        if should_be_json_body:
            headers["Content-Type"] = "application/json"

        self._add_auth_header(headers)

        params = {}
        if not url_override:
            if self.filters:
                params.update(self.filters)
            if add_params:
                params.update(add_params)

        if files:
            # Use multipart/form-data for file uploads
            req = getattr(self.http_session, verb)(
                url_override or self.url,
                headers=headers,
                params=params,
                data=data,
                files=files,
            )
        else:
            req = getattr(self.http_session, verb)(
                url_override or self.url, headers=headers, params=params, json=data
            )

        if req.status_code == 409 and verb == "post":
            raise AllocationError(req)
        if verb == "delete":
            if req.ok:
                return True
            else:
                raise RequestError(req)
        elif req.ok:
            # Parse response based on expected type
            if self.expect_json:
                try:
                    return req.json()
                except json.JSONDecodeError:
                    raise ContentError(req)
            else:
                # Return raw text for non-JSON responses
                return req.text
        else:
            raise RequestError(req)

    def concurrent_get(self, ret, page_size, page_offsets):
        futures_to_results = []
        with cf.ThreadPoolExecutor(max_workers=4) as pool:
            for offset in page_offsets:
                new_params = {"offset": offset, "limit": page_size}
                futures_to_results.append(
                    pool.submit(self._make_call, add_params=new_params)
                )

            for future in cf.as_completed(futures_to_results):
                result = future.result()
                ret.extend(result["results"])

    def get(self, add_params=None):
        """Makes a GET request.

        Makes a GET request to NetBox's API, and automatically recurses
        any paginated results.

        ## Returns
        List of `Response` objects returned from the endpoint.

        ## Raises
        * RequestError if req.ok returns false.
        * ContentError if response is not json.
        """

        if not add_params and self.limit is not None:
            add_params = {"limit": self.limit}
            if self.limit and self.offset is not None:
                # if non-zero limit and some offset -> add offset
                add_params["offset"] = self.offset
        req = self._make_call(add_params=add_params)
        if isinstance(req, dict) and req.get("results") is not None:
            self.count = req["count"]
            if self.offset is not None:
                # only yield requested page results if paginating
                for i in req["results"]:
                    yield i
            elif self.threading:
                ret = req["results"]
                if req.get("next"):
                    page_size = len(req["results"])
                    pages = calc_pages(page_size, req["count"])
                    page_offsets = [
                        increment * page_size for increment in range(1, pages)
                    ]
                    if pages == 1:
                        req = self._make_call(url_override=req.get("next"))
                        ret.extend(req["results"])
                    else:
                        self.concurrent_get(ret, page_size, page_offsets)
                for i in ret:
                    yield i
            else:
                first_run = True
                for i in req["results"]:
                    yield i
                while req["next"]:
                    # Not worrying about making sure add_params kwargs is
                    # passed in here because results from detail routes aren't
                    # paginated, thus far.
                    if first_run:
                        req = self._make_call(
                            add_params={
                                "limit": self.limit or req["count"],
                                "offset": len(req["results"]),
                            }
                        )
                    else:
                        req = self._make_call(url_override=req["next"])
                    first_run = False
                    for i in req["results"]:
                        yield i
        elif isinstance(req, list):
            self.count = len(req)
            for i in req:
                yield i
        else:
            self.count = len(req)
            yield req

    def put(self, data):
        """Makes PUT request.

        Makes a PUT request to NetBox's API.

        ## Parameters
        * **data** (dict): Contains a dict that will be turned into a
            json object and sent to the API.

        ## Returns
        Dict containing the response from NetBox's API.

        ## Raises
        * RequestError if req.ok returns false.
        * ContentError if response is not json.
        """
        return self._make_call(verb="put", data=data)

    def post(self, data):
        """Makes POST request.

        Makes a POST request to NetBox's API.

        ## Parameters
        * **data** (dict): Contains a dict that will be turned into a
            json object and sent to the API.

        ## Returns
        Dict containing the response from NetBox's API.

        ## Raises
        * RequestError if req.ok returns false.
        * AllocationError if req.status_code is 409 (Conflict)
            as with available-ips and available-prefixes when there is
            no room for the requested allocation.
        * ContentError if response is not json.
        """
        return self._make_call(verb="post", data=data)

    def delete(self, data=None):
        """Makes DELETE request.

        Makes a DELETE request to NetBox's API.

        ## Parameters
        * **data** (list): Contains a dict that will be turned into a
            json object and sent to the API.

        ## Returns
        True if successful.

        ## Raises
        RequestError if req.ok doesn't return True.
        """
        return self._make_call(verb="delete", data=data)

    def patch(self, data):
        """Makes PATCH request.

        Makes a PATCH request to NetBox's API.

        ## Parameters
        * **data** (dict): Contains a dict that will be turned into a
            json object and sent to the API.

        ## Returns
        Dict containing the response from NetBox's API.

        ## Raises
        * RequestError if req.ok returns false.
        * ContentError if response is not json.
        """
        return self._make_call(verb="patch", data=data)

    def options(self):
        """Makes an OPTIONS request.

        Makes an OPTIONS request to NetBox's API.

        ## Returns
        Dict containing the response from NetBox's API.

        ## Raises
        * RequestError if req.ok returns false.
        * ContentError if response is not json.
        """
        return self._make_call(verb="options")

    def get_count(self, *args, **kwargs):
        """Returns object count for query.

        Makes a query to the endpoint with ``limit=1`` set and only
        returns the value of the "count" field.

        ## Returns
        Int of number of objects query returned.

        ## Raises
        * RequestError if req.ok returns false.
        * ContentError if response is not json.
        """

        if not hasattr(self, "count"):
            self.count = self._make_call(add_params={"limit": 1, "brief": 1})["count"]
        return self.count



"""
(c) 2017 DigitalOcean

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from pynetbox.core.query import Request, RequestError, ParameterValidationError
from pynetbox.core.response import Record, RecordSet

RESERVED_KWARGS = ()


class Endpoint:
    """Represent actions available on endpoints in the Netbox API.

    Takes ``name`` and ``app`` passed from App() and builds the correct
    url to make queries to and the proper Response object to return
    results in.

    ## Parameters

    * **api** (Api): Takes Api created at instantiation.
    * **app** (App): Takes App.
    * **name** (str): Name of endpoint passed to App().
    * **model** (obj, optional): Custom model for given app.

    ## Note

    In order to call NetBox endpoints with dashes in their
    names you should convert the dash to an underscore.
    (E.g. querying the ip-addresses endpoint is done with
    ``nb.ipam.ip_addresses.all()``.)
    """

    def __init__(self, api, app, name, model=None):
        self.return_obj = self._lookup_ret_obj(name, model)
        self.name = name.replace("_", "-")
        self.api = api
        self.app = app
        self.base_url = api.base_url
        self.token = api.token
        self.url = "{base_url}/{app}/{endpoint}".format(
            base_url=self.base_url,
            app=app.name,
            endpoint=self.name,
        )
        self._choices = None

    def _lookup_ret_obj(self, name, model):
        """Loads unique Response objects.

        This method loads a unique response object for an endpoint if
        it exists. Otherwise return a generic `Record` object.

        ## Parameters

        * **name** (str): Endpoint name.
        * **model** (obj): The application model that contains unique Record objects.

        ## Returns
        Record (obj)
        """
        if model:
            name = name.title().replace("_", "")
            ret = getattr(model, name, Record)
        else:
            ret = Record
        return ret

    def _validate_openapi_parameters(self, method: str, parameters: dict) -> None:
        """Validate GET request parameters against OpenAPI specification

        This method raises a **ParameterValidationError** if parameters passed to NetBox API
        do not match the OpenAPI specification or validation fails.

        ## Parameters

        * **method** : Only "get" is supported as for other methods NetBox already does proper validation
        * **parameters** : kwargs passed to filter() method

        ## Returns
        None
        """
        if method.lower() != "get":
            raise RuntimeError(f"Unsupported method '{method}'.")

        openapi_definition_path = "/api/{app}/{endpoint}/".format(
            app=self.app.name,
            endpoint=self.name,
        )

        # Parse NetBox OpenAPI definition
        try:
            openapi_definition = self.api.openapi()["paths"].get(
                openapi_definition_path
            )

            if not openapi_definition:
                raise ParameterValidationError(
                    f"Path '{openapi_definition_path}' does not exist in NetBox OpenAPI specification."
                )

            openapi_parameters = openapi_definition[method]["parameters"]
            allowed_parameters = [p["name"] for p in openapi_parameters]

        except KeyError as exc:
            raise ParameterValidationError(
                f"Error while parsing Netbox OpenAPI specification: {exc}"
            )

        # Validate all parameters
        validation_errors = []
        for p in parameters:
            if p not in allowed_parameters:
                validation_errors.append(
                    f"'{p}' is not allowed as parameter on path '{openapi_definition_path}'."
                )

        if len(validation_errors) > 0:
            raise ParameterValidationError(validation_errors)

    def all(self, limit=0, offset=None):
        """Queries the 'ListView' of a given endpoint.

        Returns all objects from an endpoint.

        ## Parameters

        * **limit** (int, optional): Overrides the max page size on
            paginated returns. This defines the number of records that will
            be returned with each query to the Netbox server. The queries
            will be made as you iterate through the result set.
        * **offset** (int, optional): Overrides the offset on paginated returns.

        ## Returns
        A RecordSet object.

        ## Examples

        ```python
        devices = list(nb.dcim.devices.all())
        for device in devices:
            print(device.name)

        # test1-leaf1
        # test1-leaf2
        # test1-leaf3
        ```

        If you want to iterate over the results multiple times then
        encapsulate them in a list like this:
        ```python
        devices = list(nb.dcim.devices.all())
        ```

        This will cause the entire result set
        to be fetched from the server.
        """
        if limit == 0 and offset is not None:
            raise ValueError("offset requires a positive limit value")
        req = Request(
            base="{}/".format(self.url),
            token=self.token,
            http_session=self.api.http_session,
            threading=self.api.threading,
            limit=limit,
            offset=offset,
        )

        return RecordSet(self, req)

    def get(self, *args, **kwargs):
        """Queries the DetailsView of a given endpoint.

        ## Parameters

        * **key** (int, optional): id for the item to be retrieved.
        * **kwargs**: Accepts the same keyword args as filter(). Any search argument the endpoint accepts can
            be added as a keyword arg.
        * **strict_filters** (bool, optional): Overrides the global filter
            validation per-request basis. Handled by the filter() method.

        ## Returns
        A single Record object or None

        ## Raises
        ValueError: if kwarg search return more than one value.

        ## Examples

        Referencing with a kwarg that only returns one value:

        ```python
        nb.dcim.devices.get(name='test1-a3-tor1b')
        # test1-a3-tor1b
        ```

        Referencing with an id:

        ```python
        nb.dcim.devices.get(1)
        # test1-edge1
        ```

        Using multiple named arguments. For example, retrieving the location when the location name is not unique and used in multiple sites:

        ```python
        nb.locations.get(site='site-1', name='Row 1')
        # Row 1
        ```
        """
        try:
            key = args[0]
        except IndexError:
            key = None

        if not key:
            resp = self.filter(**kwargs)
            ret = next(resp, None)
            if not ret:
                return ret
            try:
                next(resp)
                raise ValueError(
                    "get() returned more than one result. "
                    "Check that the kwarg(s) passed are valid for this "
                    "endpoint or use filter() or all() instead."
                )
            except StopIteration:
                return ret

        req = Request(
            key=key,
            base=self.url,
            token=self.token,
            http_session=self.api.http_session,
        )
        try:
            return next(RecordSet(self, req), None)
        except RequestError as e:
            if e.req.status_code == 404:
                return None
            else:
                raise e

    def filter(self, *args, **kwargs):
        """Queries the 'ListView' of a given endpoint.

        Takes named arguments that match the usable filters on a
        given endpoint. If an argument is passed then it's used as a
        freeform search argument if the endpoint supports it.

        ## Parameters

        * **args** (str, optional): Freeform search string that's
            accepted on given endpoint.
        * **kwargs** (str, optional): Any search argument the
            endpoint accepts can be added as a keyword arg.
        * **limit** (int, optional): Overrides the max page size on
            paginated returns. This defines the number of records that will
            be returned with each query to the Netbox server. The queries
            will be made as you iterate through the result set.
        * **offset** (int, optional): Overrides the offset on paginated returns.
        * **strict_filters** (bool, optional): Overrides the global filter
            validation per-request basis.

        ## Returns
        A RecordSet object.

        ## Examples

        To return a list of objects matching a named argument filter:

        ```python
        devices = nb.dcim.devices.filter(role='leaf-switch')
        for device in devices:
            print(device.name)

        # test1-leaf1
        # test1-leaf2
        # test1-leaf3
        ```

        ```python
        devices = nb.dcim.devices.filter(site='site-1')
        for device in devices:
            print(device.name)

        # test1-a2-leaf1
        # test2-a2-leaf2
        ```

        ## Note

        If a keyword argument is incorrect a `TypeError` will not be returned by pynetbox.
        Instead, pynetbox will return all records filtered up to the last correct keyword argument. For example, if we used `site="Site 1"` instead of `site=site-1` when using filter on
        the devices endpoint, then pynetbox will return **all** devices across all sites instead of devices at Site 1.

        Using a freeform query along with a named argument:

        ```python
        devices = nb.dcim.devices.filter('a3', role='leaf-switch')
        for device in devices:
            print(device.name)

        # test1-a3-leaf1
        # test1-a3-leaf2
        ```
        """

        if args:
            kwargs.update({"q": args[0]})

        if any(i in RESERVED_KWARGS for i in kwargs):
            raise ValueError(
                "A reserved kwarg was passed ({}). Please remove it "
                "and try again.".format(RESERVED_KWARGS)
            )
        limit = kwargs.pop("limit") if "limit" in kwargs else 0
        offset = kwargs.pop("offset") if "offset" in kwargs else None
        strict_filters = (
            # kwargs value takes precedence on globally set value
            kwargs.pop("strict_filters")
            if "strict_filters" in kwargs
            else self.api.strict_filters
        )

        if limit == 0 and offset is not None:
            raise ValueError("offset requires a positive limit value")
        filters = {x: y if y is not None else "null" for x, y in kwargs.items()}

        if strict_filters:
            self._validate_openapi_parameters("get", filters)

        req = Request(
            filters=filters,
            base=self.url,
            token=self.token,
            http_session=self.api.http_session,
            threading=self.api.threading,
            limit=limit,
            offset=offset,
        )

        return RecordSet(self, req)

    def create(self, *args, **kwargs):
        """Creates an object on an endpoint.

        Takes named arguments that match the given endpoint's
        available fields. Returns a new object.

        ## Parameters

        * **args**: Not used.
        * **kwargs**: Fields and values to create the object with.

        ## Returns
        A Record object.

        ## Examples

        Creating a new device:

        ```python
        new_device = nb.dcim.devices.create(
            name='test-device',
            device_type=1,
            device_role=1,
            site=1
        )
        ```

        Creating a new device with a nested object:

        ```python
        new_device = nb.dcim.devices.create(
            name='test-device',
            device_type={'id': 1},
            device_role={'id': 1},
            site={'id': 1}
        )
        ```
        """

        req = Request(
            base=self.url,
            token=self.token,
            http_session=self.api.http_session,
        ).post(args[0] if args else kwargs)

        if isinstance(req, list):
            return [self.return_obj(i, self.api, self) for i in req]
        return self.return_obj(req, self.api, self)

    def update(self, objects):
        """Updates objects in NetBox.

        Takes a list of objects and updates them in NetBox.

        ## Parameters

        * **objects** (list): A list of Record objects to update.

        ## Returns
        A list of Record objects.

        ## Examples

        ```python
        devices = nb.dcim.devices.filter(site='test1')
        for device in devices:
            device.status = 'active'
        nb.dcim.devices.update(devices)
        ```
        """
        series = []
        if not isinstance(objects, list):
            raise ValueError(
                "Objects passed must be list[dict|Record] - was {}".format(
                    type(objects)
                )
            )
        for o in objects:
            if isinstance(o, Record):
                data = o.updates()
                if data:
                    data["id"] = o.id
                    series.append(data)
            elif isinstance(o, dict):
                if "id" not in o:
                    raise ValueError("id is missing from object: " + str(o))
                series.append(o)
            else:
                raise ValueError(
                    "Object passed must be dict|Record - was {}".format(type(objects))
                )
        req = Request(
            base=self.url,
            token=self.token,
            http_session=self.api.http_session,
        ).patch(series)

        if isinstance(req, list):
            return [self.return_obj(i, self.api, self) for i in req]
        return self.return_obj(req, self.api, self)

    def delete(self, objects):
        """Deletes objects from NetBox.

        Takes a list of objects and deletes them from NetBox.

        ## Parameters

        * **objects** (list): A list of Record objects to delete.

        ## Returns
        True if the delete operation was successful.

        ## Examples

        ```python
        devices = nb.dcim.devices.filter(site='test1')
        nb.dcim.devices.delete(devices)
        ```
        """
        cleaned_ids = []
        if not isinstance(objects, list) and not isinstance(objects, RecordSet):
            raise ValueError(
                "objects must be list[str|int|Record]"
                "|RecordSet - was " + str(type(objects))
            )
        for o in objects:
            if isinstance(o, int):
                cleaned_ids.append(o)
            elif isinstance(o, str) and o.isnumeric():
                cleaned_ids.append(int(o))
            elif isinstance(o, Record):
                if not hasattr(o, "id"):
                    raise ValueError(
                        "Record from '"
                        + o.url
                        + "' does not have an id and cannot be bulk deleted"
                    )
                cleaned_ids.append(o.id)
            else:
                raise ValueError(
                    "Invalid object in list of objects to delete: " + str(type(o))
                )

        req = Request(
            base=self.url,
            token=self.token,
            http_session=self.api.http_session,
        )
        return True if req.delete(data=[{"id": i} for i in cleaned_ids]) else False

    def choices(self):
        """Returns all choices from the endpoint if it has them.

        ## Returns
        Dictionary of available choices.

        ## Examples

        ```python
        choices = nb.dcim.devices.choices()
        print(choices['status'])
        {
            'label': 'Active',
            'value': 'active'
        }
        ```
        """
        if self._choices:
            return self._choices

        req = Request(
            base=self.url,
            token=self.api.token,
            http_session=self.api.http_session,
        ).options()

        actions = req.get("actions", {})
        post_data = actions.get("POST") or actions.get("PUT")
        if post_data is None:
            raise ValueError(
                "Unexpected format in the OPTIONS response at {}".format(self.url)
            )
        self._choices = {}
        for prop in post_data:
            if "choices" in post_data[prop]:
                self._choices[prop] = post_data[prop]["choices"]

        return self._choices

    def count(self, *args, **kwargs):
        """Returns the count of objects in a query.

        Takes named arguments that match the usable filters on a
        given endpoint. If an argument is passed then it's used as a
        freeform search argument if the endpoint supports it.

        ## Parameters

        * **args** (str, optional): Freeform search string that's
            accepted on given endpoint.
        * **kwargs** (str, optional): Any search argument the
            endpoint accepts can be added as a keyword arg.

        ## Returns
        Integer of count of objects.

        ## Examples

        ```python
        nb.dcim.devices.count(site='test1')
        # 27
        ```
        """

        if args:
            kwargs.update({"q": args[0]})

        if any(i in RESERVED_KWARGS for i in kwargs):
            raise ValueError(
                "A reserved {} kwarg was passed. Please remove it "
                "try again.".format(RESERVED_KWARGS)
            )

        ret = Request(
            filters=kwargs,
            base=self.url,
            token=self.token,
            http_session=self.api.http_session,
        )

        return ret.get_count()


class DetailEndpoint:
    """Enables read/write operations on detail endpoints.

    Endpoints like `available-ips` that are detail routes off
    traditional endpoints are handled with this class.
    """

    def __init__(self, parent_obj, name, custom_return=None):
        self.parent_obj = parent_obj
        self.custom_return = custom_return
        self.url = "{}/{}/{}/".format(parent_obj.endpoint.url, parent_obj.id, name)
        self.request_kwargs = dict(
            base=self.url,
            token=parent_obj.api.token,
            http_session=parent_obj.api.http_session,
        )

    def list(self, **kwargs):
        """The view operation for a detail endpoint.

        Returns the response from NetBox for a detail endpoint.

        ## Parameters

        * **kwargs**: Key/value pairs that get converted into URL
            parameters when passed to the endpoint.
            E.g. `.list(method='get_facts')` would be converted to
            `.../?method=get_facts`.

        ## Returns
        A Record object or list of Record objects created
        from data retrieved from NetBox.
        """
        req = Request(**self.request_kwargs).get(add_params=kwargs)

        if self.custom_return:
            return [
                self.custom_return(
                    i, self.parent_obj.endpoint.api, self.parent_obj.endpoint
                )
                for i in req
            ]
        return req

    def create(self, data=None):
        """The write operation for a detail endpoint.

        Creates objects on a detail endpoint in NetBox.

        ## Parameters

        * **data** (dict/list, optional): A dictionary containing the
            key/value pair of the items you're creating on the parent
            object. Defaults to empty dict which will create a single
            item with default values.

        ## Returns
        A Record object or list of Record objects created
        from data created in NetBox.
        """
        data = data or {}
        req = Request(**self.request_kwargs).post(data)
        if self.custom_return:
            if isinstance(req, list):
                return [
                    self.custom_return(
                        req_item, self.parent_obj.endpoint.api, self.parent_obj.endpoint
                    )
                    for req_item in req
                ]
            else:
                return self.custom_return(
                    req, self.parent_obj.endpoint.api, self.parent_obj.endpoint
                )
        return req


class RODetailEndpoint(DetailEndpoint):
    def create(self, data):
        raise NotImplementedError("Writes are not supported for this endpoint.")


class ROMultiFormatDetailEndpoint(RODetailEndpoint):
    """Read-only detail endpoint supporting multiple response formats.

    Handles endpoints that return data in different formats based on
    query parameters. Supports both structured data (JSON) and raw formats
    (e.g., SVG).

    The endpoint inspects the 'render' parameter to determine response format:
    - No parameter or render='json': Returns structured JSON data
    - render='svg': Returns raw SVG content

    ## Examples

    ```python
    rack = nb.dcim.racks.get(123)
    rack.elevation.list()  # Returns: list of rack unit objects
    rack.elevation.list(render='svg')  # Returns: SVG string
    rack.elevation.list(render='json')  # Returns: list of rack unit objects
    ```
    """

    def list(self, **kwargs):
        """Returns data in the requested format.

        ## Parameters

        * **kwargs**: Key/value pairs that get converted into URL
            parameters. Supports 'render' parameter for format selection.

        ## Returns

        - If render is non-JSON format: Raw content (string)
        - If render is 'json' or absent: Structured data (list/generator)
        """
        # Check if non-JSON format requested
        render_format = kwargs.get("render")
        if render_format == "svg":
            # Pass expect_json=False for raw SVG response
            req = Request(**self.request_kwargs, expect_json=False).get(
                add_params=kwargs
            )
            # Return raw content for non-JSON formats
            return next(req)

        if render_format != "json" and render_format is not None:
            raise ValueError(f"Unsupported render format: {render_format}")

        # Return structured JSON response via parent class
        return super().list(**kwargs)
